/*

  stacker.ino - software for an Arduino-based focus stacking and camera controller
  Written by Rylee Isitt
  Rev. 11 June 2013

  This softare is freeware, although I will not turn away donations ;)

*/

#include <EEPROM.h>
#include "nokiaLCD.h"
#include "shiftRegister8.h"

// ----------------------------------------------------------------------------------------------------
// The values of the variables below can be modified to configure the software for different hardware
// ----------------------------------------------------------------------------------------------------

// pinout configuration
// recommended pinout for the Arduino UNO
#define PIN_IR 2
#define PIN_DATA 3
#define PIN_LATCH 4
#define PIN_CLOCK 5
#define PIN_LCD_SCE   8
#define PIN_LCD_RESET 9
#define PIN_LCD_DC    10
#define PIN_LCD_SDIN  11
#define PIN_LCD_SCLK  12
#define PIN_LCD_LED 13

// motor configuration
// many of these parameters can be adjusted when the program is running, and new values will be saved in EEPROM
boolean halfStepMode = false; // toggles half-stepping
boolean reverseMotor = false; // changes the direction of the stepper motion
boolean relaxMotor = false; // if set to false, motor is always on (enables holding torque). If set to true, motor turns off between steps - this reduces torque but can prevent overheating.
unsigned int stepsPerRev = 100; // obtain this value from the motor specification (full steps per revolution, or divide 360 degrees by the number of degrees per step)
unsigned int umPerRev = 200; // to figure out this number, advance the platform through many revolutions, measure the distance moved with calipers or a micrometer, divide this distance by the number of revolutions, and round to the nearest whole number.
unsigned int maxRPM = 60; // max speed to run the motor at
unsigned int energizeTime = 5; // time (in milliseconds) to energize coils before turning them off when relaxMotor is set to true

// configuration for remote
// This project expects the Kenwood RC-P400 remote, available cheaply on eBay from multiple sellers
// If using a different remote, you will have to decipher its protocol and reprogram this sketch
// Not all remotes are equal!
// When choosing a remote, it is best if it is NOT programmable. You want buttons to always produce the same, single code using a single protocol.
// Also make sure any remote you choose has enough buttons. I recommend remotes for CD players, or other remotes designed to control a single device.
// Idea for your own ponderings: adapt this with a radio controller instead
#define REMOTE_INTERRUPT_ON FALLING // can be rising or falling
#define REMOTE_BITS 32 // the number of bits in the IR remote button signal
#define REMOTE_HIGHMAX 2500UL // elapsed time (in microseconds) since last down-sloping IR signal edge above which the code reading will reset
#define REMOTE_HIGHMIN 1500UL // elapsed time (in microseconds) since last down-sloping IR signal edge that acts as a theshhold between a 0 and 1 bit
#define REMOTE_RELEASEDELAY 100000L // elapsed time (in microseconds) after which if no bits have been received we assume that no button is being pressed
// IR codes for buttons
#define BT_NONE 0 // not a real button - a placeholder code for when no buttons are being pressed
#define BT_1 1838317695 // 1
#define BT_2 1838301375 // 2
#define BT_3 1838334015 // 3
#define BT_4 1838293215 // 4
#define BT_5 1838325855 // 5
#define BT_6 1838309535 // 6
#define BT_7 1838342175 // 7
#define BT_8 1838289135 // 8
#define BT_9 1838321775 // 9
#define BT_0 1838285055 // 0
#define BT_ENTER 1838305455 // +10 on Kenwood remote
#define BT_LEFT 1838311575 // Rewind on Kenwood remote
#define BT_RIGHT 1838344215 // Fast Forward on Kenwood remote
#define BT_UP 1838322285 // Volume Up on Kenwood remote
#define BT_DOWN 1838289645 // Volume Down on Kenwood remote
#define BT_STOP 1838319735 // Stop on Kenwood remote
#define BT_PLAYPAUSE 1838336055 // Play/pause on Kenwood remote
#define BT_MODE 1838348295 // P. Mode on Kenwood remote
#define BT_HALFPRESS 1838285820 // AI AUTO on Kenwood remote
#define BT_FULLPRESS 1838302140 // AI TIMER on Kenwood remote
#define BT_BACKLIGHT 1838318460 // PRESENCE on Kenwood remote

// shift register bitmasks
// If you change the pinout for the shift register, just modify the constants below
const byte stepperEnableA = B00000001; // shift register output pin QA/Q0
const byte stepperEnableB = B00000010; // shift register output pin QB/Q1
const byte stepperA1 = B00000100; // shift register output pin QC/Q2
const byte stepperA2 = B00001000; // shift register output pin QD/Q3
const byte stepperB1 = B00010000; // shift register output pin QE/Q4
const byte stepperB2 = B00100000; // shift register output pin QF/Q5
const byte shutterHalf = B01000000; // shift register output pin QG/Q6
const byte shutterFull = B10000000; // shift register output pin QH/Q7

// ----------------------------------------------------------------------------------------------------
// Stepper motor phase states - no need to modify these unless you want to use different excitation modes
// ----------------------------------------------------------------------------------------------------

// two-phase on, full step coil states
//   A1 B1 A2 B2
// 1 1  1  0  0
// 2 1  1  0  0
// 3 0  1  1  0
// 4 0  1  1  0
// 5 0  0  1  1
// 6 0  0  1  1
// 7 1  0  0  1
// 8 1  0  0  1
// at peak current all of the time, more heat, more torque
// makes the most out of a weak motor
const byte coilA1StateFull = B11000011;
const byte coilA2StateFull = B00111100;
const byte coilB1StateFull = B11110000;
const byte coilB2StateFull = B00001111;

// half-step coil states
//   A1 B1 A2 B2
// 1 1  0  0  0
// 2 1  1  0  0
// 3 0  1  0  0
// 4 0  1  1  0
// 5 0  0  1  0
// 6 0  0  1  1
// 7 0  0  0  1
// 8 1  0  0  1
// at peak current 1/2 of the time, less heat, less torque
// good setting for more powerful motors
const byte coilA1StateHalf = B11000001;
const byte coilA2StateHalf = B00011100;
const byte coilB1StateHalf = B01110000;
const byte coilB2StateHalf = B00000111;

// ----------------------------------------------------------------------------------------------------
// Tracking variables - no need to modify these unless you are changing the software code
// ----------------------------------------------------------------------------------------------------

// instantiate LCD
nokiaLCD lcd;

// instantiate shift register
shiftRegister8 shiftRegister;

// variables for getting IR codes
// if using a remote that uses 16 bits or less for its code
// change the variable type of code and lastCode to unsigned int
volatile unsigned long code = BT_NONE;
volatile unsigned long lastCode = BT_NONE;
volatile byte bitCount = 0;
volatile unsigned long lastBitTimestamp = 0;
volatile unsigned long curTime = 0;
volatile unsigned long elapsed = 0;

// variables for tracking button states
boolean pressState = false; // true when the button indicated by lastCode is currently held down

// variables for stepping
long stepperPos = 0; // measured in half steps, even if using full step mode
long startPos = 0; // measured in half steps, even if using full step mode
long endPos = 0; // measured in half steps, even if using full step mode


// ----------------------------------------------------------------------------------------------------
// This function loads config values from EEPROM
// ----------------------------------------------------------------------------------------------------

void loadConfig() {

  // read first 2 bytes of EEPROM - if they are not set to 64 and 128 respectively, assume that the defaults have never been written in EEPROM
  byte checkA = EEPROM.read(0);
  byte checkB = EEPROM.read(1);

  if ((checkA == 64) && (checkB == 128)) { // EEPROM values have been set
    // load values from EEPROM
    halfStepMode = (EEPROM.read(2) > 0);
    reverseMotor = (EEPROM.read(3) > 0);
    stepsPerRev = ((unsigned int)EEPROM.read(5) << 8) + EEPROM.read(4);
    umPerRev = ((unsigned int)EEPROM.read(7) << 8) + EEPROM.read(6);
    maxRPM = ((unsigned int)EEPROM.read(9) << 8) + EEPROM.read(8);
    relaxMotor = (EEPROM.read(10) > 0);
  } else { // EEPROM values have never been set
    // write current defaults to EEPROM
    EEPROM.write(0, 64);
    EEPROM.write(1, 128);
    saveConfig();
  }

}

// ----------------------------------------------------------------------------------------------------
// This function saves config values to EEPROM
// ----------------------------------------------------------------------------------------------------

void saveConfig() {
  const unsigned int firstByteMask = 255;
  const unsigned int secondByteMask = 65280;
  EEPROM.write(2, halfStepMode);
  EEPROM.write(3, reverseMotor);
  EEPROM.write(4, stepsPerRev & firstByteMask); // first (rightmost) byte
  EEPROM.write(5, (stepsPerRev & secondByteMask) >> 8); // second (leftmost) byte
  EEPROM.write(6, umPerRev & firstByteMask); // first (rightmost) byte
  EEPROM.write(7, (umPerRev & secondByteMask) >> 8); // second (leftmost) byte
  EEPROM.write(8, maxRPM & firstByteMask); // first (rightmost) byte
  EEPROM.write(9, (maxRPM & secondByteMask) >> 8); // second (leftmost) byte
  EEPROM.write(10, relaxMotor);
}

// ----------------------------------------------------------------------------------------------------
// This function is called by an interrupt when the IR receiver breakout board signal falls
// It measures the intervals between falling edges and uses these intervals to decode the IR signal
// ----------------------------------------------------------------------------------------------------

void irFire() {
  /*
    Deciphering codes from the remote works by watching the elapsed time between down-sloping or up-sloping edges (as a high signal transitions to a low signal or vice versa)
    When this happens, irFire() is triggered by an interrupt
    ```|___|````|____
       a        b
    In the waveform above, this elapsed time is the period between "a" and "b"
    If the elapsed time is between REMOTE_HIGHMIN and REMOTE_HIGHMAX, this is interpreted as a 1
    If the elapsed time is below REMOTE_HIGHMIN, this is interpreted as a 0
    If the elapsed time is above REMOTE_HIGHMAX, this resets code reading (sets code to 0 and bitCount to 0)
    Once REMOTE_BITS bits have been received, the code is stored in the lastCode variable
  */

  curTime = micros();
  elapsed = curTime - lastBitTimestamp;
  lastBitTimestamp = curTime;

  // watch the bits coming from the receiver
  bitCount++;
  if (elapsed < REMOTE_HIGHMIN) { // zero
    code <<= 1;
  } else if (elapsed < REMOTE_HIGHMAX) { // one
    code <<= 1;
    code++;
  } else { // begin a new code
    code = 0;
    bitCount = 0;
  }

  if (bitCount == REMOTE_BITS) { // a complete code has been received received
    lastCode = code;
    bitCount = 0;
    code = 0;
  }

}

// ----------------------------------------------------------------------------------------------------
// This function moves the stepper by one step (or one half-step) in either direction
// ----------------------------------------------------------------------------------------------------

void singleStep(boolean forward, boolean halfMode) {

  static byte stepperStageBitMask = B10000000;

  // move stepperStageBitMask forwards or backwards
  if ((forward && !reverseMotor) || (!forward && reverseMotor)) {
    stepperStageBitMask = stepperStageBitMask >> 1;
    if (stepperStageBitMask == 0) stepperStageBitMask = B10000000;
    stepperPos++;
  } else {
    stepperStageBitMask = stepperStageBitMask << 1;
    if (stepperStageBitMask == 0) stepperStageBitMask = B00000001;
    stepperPos--;
  }

  if (halfMode) {
    shiftRegister.set(stepperEnableA, ((stepperStageBitMask & coilA1StateHalf) > 0) || ((stepperStageBitMask & coilA2StateHalf) > 0));
    shiftRegister.set(stepperEnableB, ((stepperStageBitMask & coilB1StateHalf) > 0) || ((stepperStageBitMask & coilB2StateHalf) > 0));
    shiftRegister.set(stepperA1, (stepperStageBitMask & coilA1StateHalf) > 0);
    shiftRegister.set(stepperA2, (stepperStageBitMask & coilA2StateHalf) > 0);
    shiftRegister.set(stepperB1, (stepperStageBitMask & coilB1StateHalf) > 0);
    shiftRegister.set(stepperB2, (stepperStageBitMask & coilB2StateHalf) > 0);
  } else {
    shiftRegister.set(stepperEnableA, ((stepperStageBitMask & coilA1StateFull) > 0) || ((stepperStageBitMask & coilA2StateFull) > 0));
    shiftRegister.set(stepperEnableB, ((stepperStageBitMask & coilB1StateFull) > 0) || ((stepperStageBitMask & coilB2StateFull) > 0));
    shiftRegister.set(stepperA1, (stepperStageBitMask & coilA1StateFull) > 0);
    shiftRegister.set(stepperA2, (stepperStageBitMask & coilA2StateFull) > 0);
    shiftRegister.set(stepperB1, (stepperStageBitMask & coilB1StateFull) > 0);
    shiftRegister.set(stepperB2, (stepperStageBitMask & coilB2StateFull) > 0);
  }
  
  if (relaxMotor) {
    // disable the stepper after a short time
    delay(energizeTime);
    shiftRegister.set(stepperEnableA, false);
    shiftRegister.set(stepperEnableB, false);
  }

}

// ----------------------------------------------------------------------------------------------------
// When continuously stepping, this function checks to see if it's time to step again, and if so, makes
// a single step (or half-step)
// ----------------------------------------------------------------------------------------------------

void multiStep(boolean forward, boolean halfMode) {

  // timing
  static unsigned long lastStepTime = 0;
  unsigned long curMillis = millis();

  // movement
  if (curMillis >= (lastStepTime + (60000 / (maxRPM * stepsPerRev * 2)))) {
    lastStepTime = curMillis;
    singleStep(forward, halfMode);
  }

}

// ----------------------------------------------------------------------------------------------------
// This function rotates the stepper to a destination position in either full-step or half-step mode
// ----------------------------------------------------------------------------------------------------

void gotoStep(int goalPos, boolean halfMode) {

  // go to goalPos
  while (stepperPos != goalPos) {
    if (((stepperPos < goalPos) && !reverseMotor) || ((stepperPos > goalPos) && reverseMotor)) {
      multiStep(true,halfMode);
    } else {
      multiStep(false,halfMode);
    }
  }

}

// ----------------------------------------------------------------------------------------------------
// This function sets the button state based on the received IR codes and the timing of the last bit
// ----------------------------------------------------------------------------------------------------

void handleButtons(void) {
  
  // pressState == true when a button is held down or was recently released
  // the lag in determinining if a button has been released depends on REMOTE_RELEASEDELAY
  // this lag is due to the delay between signal bits from the remote
  
  if (!pressState) lastCode = BT_NONE;
  pressState = (((long)micros() - (long)lastBitTimestamp) < REMOTE_RELEASEDELAY);

}

// ----------------------------------------------------------------------------------------------------
// This function returns a byte corresponding to the digit pressed on the remote
// 255 is returned if no digit button has been pressed
// ----------------------------------------------------------------------------------------------------

byte lastRemoteDigit(void) {
  byte remoteDigit = 255;
  if (!pressState) {
    if (lastCode == BT_1) remoteDigit = 1;
    else if (lastCode == BT_2) remoteDigit = 2;
    else if (lastCode == BT_3) remoteDigit = 3;
    else if (lastCode == BT_4) remoteDigit = 4;
    else if (lastCode == BT_5) remoteDigit = 5;
    else if (lastCode == BT_6) remoteDigit = 6;
    else if (lastCode == BT_7) remoteDigit = 7;
    else if (lastCode == BT_8) remoteDigit = 8;
    else if (lastCode == BT_9) remoteDigit = 9;
    else if (lastCode == BT_0) remoteDigit = 0;
  }
  return remoteDigit;
}

// ----------------------------------------------------------------------------------------------------
// This function adds digits to the end (rightmost) of an unsigned integer, and prevents overflow
// ----------------------------------------------------------------------------------------------------

boolean addDigitUI(unsigned int &param, byte digit, boolean appendMode) {
  if (appendMode) {
    if ((param <= 6552) || ((param == 6553) && (digit <= 5))) {
      param = (param * 10) + digit;
    } else {
      param = 65536;
    }
  } else {
    param = digit;
  }
  return true;
}

// ----------------------------------------------------------------------------------------------------
// This function handles the configuration menu
// ----------------------------------------------------------------------------------------------------

void config(void) {

  // the configuration allows to user to set:
  //   - halfStepMode
  //   - reverseMotor
  //   - stepsPerRev
  //   - umPerRev
  //   - maxRPM

  byte curParam = 0;
  boolean appendMode = false;

  while ((lastCode != BT_MODE) || pressState) {

    // handle interpretation of codes received from remote
    handleButtons();
    byte remoteDigit = lastRemoteDigit();

    if ((lastCode == BT_BACKLIGHT) && !pressState) {
      lcd.toggleBacklight(); // backlight toggle
    } else if ((lastCode == BT_UP) && !pressState) { // cycle selected field up
      if (curParam == 0) curParam = 6;
      curParam--;
      appendMode = false;
    } else if ((lastCode == BT_DOWN) && !pressState) { // cycle selected field down
      curParam++;
      if (curParam >= 6) curParam = 0;
      appendMode = false;
    } else if (remoteDigit != 255) {
      if (curParam==0) halfStepMode = (remoteDigit > 0);
      if (curParam==1) reverseMotor = (remoteDigit > 0);
      if (curParam==2) relaxMotor = (remoteDigit > 0);
      if (curParam==3) appendMode = addDigitUI(stepsPerRev,remoteDigit,appendMode);
      if (curParam==4) appendMode = addDigitUI(umPerRev,remoteDigit,appendMode);
      if (curParam==5) appendMode = addDigitUI(maxRPM,remoteDigit,appendMode);
    } else {
      // update display
      lcd.cls();
      lcd.writeLine("Half-step:"+String(halfStepMode),(curParam==0));
      lcd.writeLine("Reverse:"+String(reverseMotor),(curParam==1));
      lcd.writeLine("Relax motor:"+String(relaxMotor),(curParam==2));
      lcd.writeLine("St/rev:"+String(stepsPerRev),(curParam==3));
      lcd.writeLine("um/rev:"+String(umPerRev),(curParam==4));
      lcd.writeLine("Max RPM:"+String(maxRPM),(curParam==5));
      lcd.update();
    }

  }

  // enforce minimum values
  if (stepsPerRev < 4) stepsPerRev = 4; // this is 90 degrees per step
  if (umPerRev < 1) umPerRev = 1;
  if (maxRPM < 1) maxRPM = 1;

  // store the updated parameters in EEPROM
  saveConfig();

}

// ----------------------------------------------------------------------------------------------------
// This function performs the stacking
// ----------------------------------------------------------------------------------------------------

void doStack(unsigned int slices, unsigned int interval, unsigned int bracket, unsigned int exposureTime) {

  // update display
  lcd.cls();
  lcd.writeLine("---Stacking---",true);
  lcd.writeLine("Press stop");
  lcd.writeLine("To cancel.");
  lcd.writeLine("!\"\"\"\"\"\"\"\"\"\"\"\"#",false,true);
  lcd.update();

  // stack
  for (int i=0; i<slices; i++) {
    unsigned int waitHalfTime = (interval - exposureTime) / 2;

    // calculate target position
    int gotoPos = startPos + ((long(i)*(long(endPos-startPos))) / (int(slices) - 1));

    // update progress bar
    byte progress = ((i*14) / (slices-1));
    for (byte j=0; j<progress; j++) {
      lcd.setRow(3);
      lcd.setCol(j);
      if (j == 0) {
        lcd.writeLine("$",false,true);
      } else if (j == 13) {
        lcd.writeLine("&",false,true);
      } else {
        lcd.writeLine("%",false,true);
      }
    }
    lcd.update();

    // go to target position
    gotoStep(gotoPos, halfStepMode);

    // take picture
    for (byte n=0; n<bracket; n++) {

      // enable half-press mode
      shiftRegister.set(shutterHalf,true);

      delay(waitHalfTime); // wait (to reduce camera vibrations)

      // trigger camera shutter
      shiftRegister.set(shutterFull,true);
      delay(exposureTime);

      // release camera shutter
      shiftRegister.set(shutterHalf,false);
      shiftRegister.set(shutterFull,false);

      delay(waitHalfTime); // wait (to reduce camera vibrations)

    }

    // if stop button pressed, cancel
    handleButtons();
    if (lastCode == BT_STOP) i = slices;

  }
}

// ----------------------------------------------------------------------------------------------------
// This function handles the stack setup
// ----------------------------------------------------------------------------------------------------

void stackMode(void) {

  static unsigned int slices = 80; // step count
  static unsigned int interval = 15000; // wait time (milliseconds between slices - the default of 15000 ms is to prevent flash overheating)
  static unsigned int bracket = 1; // bracket count (number of shots to take per slice)
  static unsigned int exposureTime = 100; // time (in milliseconds) that shutter release is held down
  byte curParam = 0;
  boolean appendMode = false;

  while ((lastCode != BT_MODE) || pressState) {

    // handle interpretation of codes received from remote
    handleButtons();
    byte remoteDigit = lastRemoteDigit();

    // enforce minimum values

    if (((lastCode == BT_UP) || (lastCode == BT_DOWN) || (lastCode == BT_PLAYPAUSE)) && !pressState) {
      if (slices < 2) slices = 2;
      if (exposureTime < 50) exposureTime = 50;
      if (interval < (exposureTime  + 100)) interval = exposureTime + 100;
      if (bracket < 1) bracket = 1;
    }

    if ((lastCode == BT_BACKLIGHT) && !pressState) {
      lcd.toggleBacklight(); // backlight toggle
    } else if ((lastCode == BT_UP) && !pressState) { // cycle selected field up
      if (curParam == 0) curParam = 4;
      curParam--;
      appendMode = false;
    } else if ((lastCode == BT_DOWN) && !pressState) { // cycle selected field down
      curParam++;
      if (curParam >= 4) curParam = 0;
      appendMode = false;
    } else if (remoteDigit != 255) {
      if (curParam==0) appendMode = addDigitUI(slices,remoteDigit,appendMode);
      if (curParam==1) appendMode = addDigitUI(interval,remoteDigit,appendMode);
      if (curParam==2) appendMode = addDigitUI(bracket,remoteDigit,appendMode);
      if (curParam==3) appendMode = addDigitUI(exposureTime,remoteDigit,appendMode);
    } else if ((lastCode == BT_PLAYPAUSE) && !pressState) { // do the stack
      doStack(slices, interval, bracket, exposureTime);
    } else {
      // update display
      unsigned int stepDiff = abs(startPos - endPos);
      unsigned long distance = ((unsigned long)umPerRev * (unsigned long)stepDiff) / ((unsigned long)stepsPerRev * 2UL);
      lcd.cls();
      lcd.writeLine("Depth:"+String(distance)+"um");
      lcd.writeLine("Slices:"+String(slices),(curParam==0));
      lcd.writeLine("Int:"+String(interval)+"ms",(curParam==1));
      lcd.writeLine("Bracket:"+String(bracket),(curParam==2));
      lcd.writeLine("Exp:"+String(exposureTime)+"ms",(curParam==3));
      lcd.update();
    }
  }
}

// ----------------------------------------------------------------------------------------------------
// This function handles the UI and stepping in the position mode
// ----------------------------------------------------------------------------------------------------

void positionMode(void) {

  // mode 1: positioning - allows manual movement of rack and setting of extents
  //    up: moves rack forward
  //    down: moves rack backward
  //    left: goes to start position
  //    right: goes to end position
  //    play/pause: sets start position
  //    stop: sets end position

  boolean halfPressState = false;

  while ((lastCode != BT_MODE) || pressState) {

    // handle interpretation of codes received from remote
    handleButtons();

    // shutter
    if ((lastCode == BT_HALFPRESS) && !pressState) {
      halfPressState = !halfPressState;
      shiftRegister.set(shutterHalf,halfPressState);
    }
    if ((lastCode == BT_FULLPRESS) && pressState) {
      shiftRegister.set(shutterFull,true);
    } else {
      shiftRegister.set(shutterFull,false);
    }

    if ((lastCode == BT_BACKLIGHT) && !pressState) {
      lcd.toggleBacklight(); // backlight toggle
    } else if ((lastCode == BT_PLAYPAUSE) && !pressState) {
      startPos = stepperPos;
    } else if ((lastCode == BT_STOP) && !pressState) {
      endPos = stepperPos;
    } else if ((lastCode == BT_UP) && pressState) {
      multiStep(true, halfStepMode);
    } else if ((lastCode == BT_DOWN) && pressState) {
      multiStep(false, halfStepMode);
    } else if ((lastCode == BT_RIGHT) && !pressState) {
      gotoStep(endPos, halfStepMode);
    } else if ((lastCode == BT_LEFT) && !pressState) {
      gotoStep(startPos, halfStepMode);
    } else { // update display
      // calculate physical distancce between start and end points
      unsigned int stepDiff = abs(startPos - endPos);
      unsigned long distance = ((unsigned long)umPerRev * (unsigned long)stepDiff) / ((unsigned long)stepsPerRev * 2UL);
      lcd.cls();
      lcd.writeLine("Pos:"+String(stepperPos));
      lcd.writeLine("Start:"+String(startPos));
      lcd.writeLine("End:"+String(endPos));
      lcd.writeLine("Depth:"+String(distance)+"um");
      if (halfPressState) lcd.writeLine("*HALF PRESS*");
      lcd.update();
    }

  }

  // make sure the shutter is off
  shiftRegister.set(shutterHalf,false);
  shiftRegister.set(shutterFull,false);

}

// ----------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------

void setup(void) {

  // load configuration values from EEPROM
  loadConfig();

  // initialize LCD
  lcd.begin(PIN_LCD_SCE, PIN_LCD_RESET, PIN_LCD_DC, PIN_LCD_SDIN, PIN_LCD_SCLK, PIN_LCD_LED);

  // initialize shift register
  shiftRegister.begin(PIN_DATA, PIN_LATCH, PIN_CLOCK);

  // setup pin for IR in
  pinMode(PIN_IR, INPUT);
  digitalWrite(PIN_IR, HIGH); // pulls high
  attachInterrupt(0, irFire, REMOTE_INTERRUPT_ON);
  interrupts();

}

// ----------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------

void loop(void) {

  // handle interpretation of codes received from remote
  handleButtons();
  
  /*
  mode button cycles through the modes below:
  mode 1: positioning - allows manual movement of rack and setting of extents
  mode 2: stacking - automatically increments between start and stop positions while triggering camera shutter
  mode 3: configuration - allows user to set several adjustable parameters
  todo: interval - for timelapse, with support for a motorized mount driven by a stepper (WORK IN PROGRESS)
  todo: equatorial - for driving a barn door mount using a stepper and doing long exposures while mount is running (WORK IN PROGRESS)
  */

  // deal with buttons
  if ((lastCode == BT_BACKLIGHT) && !pressState) { // enable/disable backlight
    lcd.toggleBacklight();
  } else if ((lastCode == BT_1) && !pressState) {
    positionMode();
  } else if ((lastCode == BT_2) && !pressState) {
    stackMode();
  } else if ((lastCode == BT_3) && !pressState) {
    config();
  } else { // update display
    lcd.cls();
    lcd.writeLine("1:Position");
    lcd.writeLine("2:Stack");
    lcd.writeLine("3:Config");
    lcd.update();
  }

}
